# == Schema Information
#
# Table name: components
#
#  id                            :bigint           not null, primary key
#  delegate_requests_through_hub :boolean          default(FALSE)
#  description                   :string(4000)
#  external_address              :string(4000)
#  external_port                 :integer
#  has_token                     :boolean
#  has_update_available          :boolean
#  highest_available_update      :string(4000)
#  ip_address                    :string(4000)
#  is_api                        :boolean
#  is_updating                   :boolean
#  last_ping_at                  :datetime
#  last_ping_result              :text(2147483647)
#  last_successful_ping_at       :datetime
#  last_update_error             :string(4000)
#  licenced                      :boolean
#  module_name                   :string(4000)
#  port                          :integer
#  protocol                      :string(4000)     default("http")
#  reachable                     :boolean
#  token                         :string(4000)
#  type                          :string(4000)
#  update_status                 :string(4000)
#  version                       :string(4000)
#  created_at                    :datetime         not null
#  updated_at                    :datetime         not null
#  email_id                      :bigint
#  module_id                     :string(4000)
#
# Indexes
#
#  index_components_on_email_id             (email_id)
#  index_components_on_module_id            (module_id) UNIQUE
#  index_components_on_port_and_ip_address  (port,ip_address) UNIQUE
#
# Foreign Keys
#
#  fk_rails_...  (email_id => emails.id) ON DELETE => nullify
#
class Component < ApplicationRecord
  has_many :component_permissions
  belongs_to :email, optional: true

  validates_uniqueness_of :module_id, allow_nil: true
  validates_uniqueness_of :port, scope: :ip_address

  has_one_attached :icon_file
  has_secure_token

  class << self
    def hub_url
      @hub_url ||= find_by(module_id: "rx_hub")&.index
    end

    def display_name
      "Apps"
    end

    def ping
      ensure_self
      eager_load(:component_permissions).all.each do |component|
        component = component.becomes(component.type.constantize)

        component.licenced = if component.module_id
          RxModule::Licence.module_allowed?(component.module_id)
        else
          false
        end

        component.ping

        # only hit the database if the licenced field actually changed
        # since this runs every minute
        component.save if component.changed?
      end
    end

    def resolve(internal_id_or_module_id)
      if internal_id_or_module_id.is_a?(Integer) || /\b(?<!\.)\d+(?!\.)\b/ === internal_id_or_module_id
        begin
          find(internal_id_or_module_id)
        rescue ActiveRecord::RecordNotFound
          nil
        end
      else
        find_by(module_id: internal_id_or_module_id)
      end
    end

    def ensure_self
      RxModuleComponent.where(module_id: RxModule.module_id).first_or_create.tap do |mod|
        mod.module_name = RxModule.module_name
        if mod.ip_address.nil? || mod.ip_address.empty?
          mod.ip_address = ENV["RX_HUB_IP_ADDRESS"] || Socket.ip_address_list.detect(&:ipv4_private?).try(:ip_address) || "127.0.0.1"
        end
        mod.port = ENV["MODULE_PORT"] || (Rails.env.development? ? 3000 : 80)
        mod.save if mod.changed?
      end
    end

    def get_components_version_info
      prerelease_components = []
      stable_components = []

      begin
        if RxModuleComponent.update_server
          resp = Faraday.get("#{RxModuleComponent.update_server}/packages?prerelease=true", {}, "Accept" => "application/json")
          Rails.logger.debug { {action: "Package Check", package_server_resp: resp, body: resp.body} }

          prerelease_components = JSON.parse(resp.body)

          resp = Faraday.get("#{RxModuleComponent.update_server}/packages", {}, "Accept" => "application/json")
          stable_components = JSON.parse(resp.body)
        else
          Rails.logger.warn { {action: "Package Check", msg: "No update server configured", server: RxModuleComponent.update_server} }
        end
      rescue
        Rails.logger.warn { {action: "Package Check", msg: "Could not reach update server or response malformed", server: RxModuleComponent.update_server} }
      end
      components = prerelease_components.map do |code, versions|
        applicable_versions = if LicenceInfo.can_download_prerelease?(code)
          versions
        else
          stable_components[code]
        end
        {
          id: code,
          versions: applicable_versions || [],
          installable: RxModuleComponent.module_installable?(code)
        }
      end
      components.compact
    end

    def up_to_date_components_info
      components = get_components_version_info
      set_update_available_info(components)
      components
    end

    def set_update_available_info(update_data = nil)
      update_data = get_components_version_info if update_data.nil?
      Component.where(module_id: update_data.map { |u| u[:id] }).all.each do |component|
        this_update = update_data.find do |u|
          u[:id] == component.module_id
        end
        versions = this_update[:versions].map do |v|
          Semantic::Version.new(v)
        rescue
          Rails.logger.warn { {msg: "Error in parsing module version"} }
        end.compact
        highest_version = versions.max
        component.highest_available_update = highest_version
        component.save if component.changed?
      rescue
        Rails.logger.error { {msg: "Issue setting update data", component: component} }
      end
    end
  end

  compute_on_save(:has_update_available) do |c|
    !c.up_to_date?
  end

  def up_to_date?
    return true if highest_available_update.nil? || version.nil?
    Semantic::Version.new(highest_available_update) <= Semantic::Version.new(version)
  end

  def upgrade!(target_version = nil)
    raise "No generic upgrade implemented"
  end

  def to_json(*args)
    render_jbuilder_partial
  rescue
    super(*args)
  end

  def index
    "#{protocol || "http"}://#{ip_address}:#{port}"
  end

  def external_index
    if external_address.present?
      "#{protocol || "http"}://#{external_address}:#{external_port}"
    else
      index
    end
  end

  def icon_url
    "#{Component.hub_url}/api/components/#{id}/icon?v=#{version}"
  end

  def controller_name
    "ComponentsController"
  end

  def broadcast_class_name
    "RxModuleComponent"
  end

  def broadcast_local_name
    :component
  end

  def call_module(endpoint, method: :get, params: {}, headers: {})
    RxModule::Communicator.call_module_method(module_id, endpoint, method, params, headers)
  end

  def resolve_endpoint(endpoint)
    return index if endpoint.nil? || endpoint.empty?
    endpoint = endpoint.split("/").reject(&:empty?)
    "#{index}/#{endpoint.join("/")}"
  end

  def resolve_endpoint_external(endpoint)
    return external_index if endpoint.blank?
    endpoint = endpoint.split("/").reject(&:empty?)
    "#{external_index}/#{endpoint.join("/")}"
  end

  def endpoint_authorization(endpoint)
    RxModule::Authorization.current_user_token
  end

  def mark_downloading_update
    update(is_updating: true, update_status: "Downloading package")
  end

  def mark_transferring_update
    update(is_updating: true, update_status: "Transferring package")
  end

  def mark_update_error(details)
    update(is_updating: false, update_status: nil, last_update_error: details)
  end

  def mark_update_applying
    update(reachable: false, last_update_error: nil, update_status: "Applying package")
  end

  def mark_update_step(msg)
    update(update_status: msg)
  end
end
