# == Schema Information
#
# Table name: components
#
#  id                            :bigint           not null, primary key
#  delegate_requests_through_hub :boolean          default(FALSE)
#  description                   :string(4000)
#  external_address              :string(4000)
#  external_port                 :integer
#  has_token                     :boolean
#  has_update_available          :boolean
#  highest_available_update      :string(4000)
#  ip_address                    :string(4000)
#  is_api                        :boolean
#  is_updating                   :boolean
#  last_ping_at                  :datetime
#  last_ping_result              :text(2147483647)
#  last_successful_ping_at       :datetime
#  last_update_error             :string(4000)
#  licenced                      :boolean
#  module_name                   :string(4000)
#  port                          :integer
#  protocol                      :string(4000)     default("http")
#  reachable                     :boolean
#  token                         :string(4000)
#  type                          :string(4000)
#  update_status                 :string(4000)
#  version                       :string(4000)
#  created_at                    :datetime         not null
#  updated_at                    :datetime         not null
#  email_id                      :bigint
#  module_id                     :string(4000)
#
# Indexes
#
#  index_components_on_email_id             (email_id)
#  index_components_on_module_id            (module_id) UNIQUE
#  index_components_on_port_and_ip_address  (port,ip_address) UNIQUE
#
# Foreign Keys
#
#  fk_rails_...  (email_id => emails.id) ON DELETE => nullify
#
class RxModuleComponent < Component
  class << self
    def update_server
      ENV["UPDATE_SERVER"] || RxModule::Licence.licence_server
    end

    # anything the customer is allowed to install (by licence)
    def eligible_installations
      %w[rx_pdf]
    end

    # allowed installations - performed installations
    def available_installations
      eligible_installations.reject do |module_code|
        module_installed?(module_code)
      end
    end

    def module_installed?(module_code)
      exists?(module_id: module_code)
    end

    def module_installable?(module_code)
      available_installations.include?(module_code.to_s)
    end
  end

  def to_partial_path
    Component._to_partial_path
  end

  def upgrade!(version = nil)
    mark_downloading_update

    url = "#{RxModuleComponent.update_server}/package/#{module_id}"
    query = {}
    query[:prerelease] = true if LicenceInfo.can_download_prerelease?(module_id)
    query[:version] = version if version.present?

    url = "#{url}?#{query.to_query}"

    client = Faraday.new(url: url) do |faraday|
      faraday.use FaradayMiddleware::FollowRedirects
      faraday.adapter Faraday.default_adapter
    end
    response = client.get(url)
    if response.status == 200
      begin
        mark_transferring_update
        Tempfile.create(binmode: true) do |file|
          file.write(response.body)
          conn = Faraday.new(nil, ssl: {
            verify: !RxModule::Communicator.skip_ssl_validation
          }) do |f|
            f.request :multipart
          end
          file_part = Faraday::FilePart.new(file.path, "application/zip")
          update_resp = conn.post(resolve_endpoint("/api/rx_module/update")) do |req|
            req.headers[:Authorization] = "TOKEN #{RxModule::Authorization.current_user_token}"
            req.body = {file: file_part}
          end
          update_json = JSON.parse(update_resp.body)
          Rails.logger.info { {action: "Update response", resp: update_json, component: self} }
          if update_json["error"]
            mark_update_error(update_json["error_details"])
            ping
            raise update_json["error_details"]
          else
            mark_update_applying
          end
          file_part.io.close
          file.close
        end
        update(has_update_available: false)
      rescue Errno::EACCES
        Rails.logger.warn { "Failed to delete tmpfile" }
      end
    else
      mark_update_error("Unreachable update server")
      raise "Unreachable update server"
    end
  end

  def update_icon
    if is_hub?
      icon = File.open(RxModule.module_icon)
      icon_file.attach(io: icon, filename: File.basename(icon.path))
    else
      conn = Faraday.new(resolve_endpoint("/api/rx_module/icon"), ssl: {
        verify: !RxModule::Communicator.skip_ssl_validation
      })
      svg_resp = conn.get(nil)
      Tempfile.create(%w[icon_download .svg]) do |f|
        f.write(svg_resp.body)
        f.flush
        f.rewind
        icon_file.attach(io: f, filename: File.basename(f.path), content_type: svg_resp.headers["content-type"], identify: false)
      end
    end

  end

  def update_token
    if is_hub?
      RxModule::Token.set token
    else
      RxModule::Communicator.call_module_method(module_id, "/api/rx_module/token", :put, {
        token: token
      })
    end
  end

  def update_owner
    if is_hub?
      RxModule::Token.set_owner SystemSetting.instance.is_satisloh_customer? ? 'SATISLOH' : 'RXUN'
    else
      RxModule::Communicator.call_module_method(module_id, "/api/rx_module/owner", :put, {
        owner: SystemSetting.instance.is_satisloh_customer? ? 'SATISLOH' : 'RXUN'
      })
    end
  end

  def is_hub?
    module_id == RxModule.module_id
  end

  def get_info
    if is_hub?
      RxModule.info
    else
      RxModule::Communicator.call_module_method(module_id, "/api/rx_module/info", :get) do |req|
        req.options.timeout = 5 # info should be *fast* or we can consider it unreachable
      end
    end
  end

  def healthcheck
    if is_hub?
      true
    else
      RxModule::Communicator.call_module_method(module_id, "/api/rx_module/up", :get)
      true
    end
  rescue
    false
  end

  def sync(response = nil)
    old_version = version
    response = get_info if response.nil?
    process_ping_result(response)
    begin
      if reachable?
        if (old_version != version) || !icon_file.attached?
          update_icon
        end

        # this is deliberately checking false because if it's nil,
        # the module doesn't support tokens yet
        update_token if has_token == false

        # Only attempt to update the module owner if the module supports it
        if response.key?(:owner)
          update_owner if response[:owner].nil?
        end
      end
    rescue => err
      Rails.logger.warn { {msg: "Issue updating module icon", component: as_json} }
    end
  rescue => err
    error_info = {
      msg: "Component may be offline or returning invalid responses",
      component: as_json(except: :last_ping_result),
      exception: {
        msg: err.message,
        stack_trace: err.backtrace
      }
    }
    Rails.logger.info do
      error_info
    end
    assign_attributes(reachable: false, last_ping_result: error_info.to_json)
  ensure
    self.last_ping_at = Time.current
    save
  end

  # Pings the module
  #
  # If the module was offline before this ping, and is online now,
  # sync the module info
  # If the module is offline, it's either actually offline, or it's running an old rx_module version
  # Try syncing, if the module is actually offline this should have no performance impact because the request will just fail
  # If it's online, it'll sync just as it did before
  def ping
    if healthcheck
      sync unless reachable?
    else
      sync
    end
  end

  def process_ping_result(status)
    if status
      ActiveRecord::Base.transaction do
        # we restrict the updates to minimize websocket traffic.  We'll only broadcast if something is dirty
        updates = {}
        updates[:reachable] = true unless self[:reachable]
        {
          module_id: :id,
          module_name: :name,
          description: :description,
          version: :version,
          is_api: :is_api
        }.each do |k_this, k_resp|
          updates[k_this] = status[k_resp] unless self[k_this] == status[k_resp]
        end

        # If this is present, this means the module itself has support for tokens
        # If it's nil, we want to keep has_token nil, not false, so we can tell the difference
        unless status[:has_token].nil?
          updates[:has_token] = status[:has_token]
        end

        if status[:last_update]
          updates[:is_updating] = status[:last_update][:in_progress] unless self[:is_updating] == status[:last_update][:in_progress]
          updates[:last_update_error] = status[:last_update][:error_msg] unless self[:last_update_error] == status[:last_update][:error_msg]
          updates[:update_status] = nil unless updates[:is_updating]
        else
          updates[:is_updating] = false unless self[:is_updating] == false
          updates[:last_update_error] = nil unless self[:last_update_error].nil?
          updates[:update_status] = nil unless self[:update_status].nil?
        end
        if status[:ssl]
          updates[:protocol] = "https" unless self[:protocol] == "https"
        else
          updates[:protocol] = "http" unless self[:protocol] == "http"
        end

        updates[:last_ping_result] = status.to_json
        updates[:last_successful_ping_at] = Time.current

        assign_attributes(updates) if updates.size > 0

        permissions_to_revoke = []

        if status[:permissions]
          permissions_to_keep = []
          status[:permissions].each do |target_definition|
            target_definition[:actions].each do |action_definition|
              # component_permissions are eager loaded, so avoid hitting the database here
              permission = component_permissions.find { |cp| cp.target_id == target_definition[:id] && cp.action_id == action_definition[:id] }
              permission_allowed = true
              if action_definition[:feature_subfeature_id] && !RxModule::Licence.feature_allowed?(action_definition[:feature_module_id], action_definition[:feature_subfeature_id])
                permission_allowed = false
                if permission
                  Rails.logger.info { "Feature not allowed (#{permission[:feature_module_id]}, #{permission[:feature_subfeature_id]}) - revoking permission #{target_definition[:name]}:#{action_definition[:name]} from all roles" }
                  permissions_to_revoke << permission
                end
              elsif action_definition[:feature_module_id] && !RxModule::Licence.module_allowed?(action_definition[:feature_module_id])
                permission_allowed = false
                if permission
                  Rails.logger.info { "Feature not allowed (#{permission[:feature_module_id]}, #{permission[:feature_subfeature_id]}) - revoking permission #{target_definition[:name]}:#{action_definition[:name]} from all roles" }
                  permissions_to_revoke << permission
                end
              end
              if permission_allowed
                permission ||= component_permissions.new(target_id: target_definition[:id], action_id: action_definition[:id])
                updates = {}
                if permission[:target_display_name] != target_definition[:name]
                  updates[:target_display_name] = target_definition[:name]
                end
                if permission[:action_display_name] != action_definition[:name]
                  updates[:action_display_name] = action_definition[:name]
                end
                if permission[:target_category] != target_definition[:category]
                  updates[:target_category] = target_definition[:category]
                end
                if permission[:feature_module_id] != action_definition[:feature_module_id]
                  updates[:feature_module_id] = action_definition[:feature_module_id]
                end
                if permission[:feature_subfeature_id] != action_definition[:feature_subfeature_id]
                  updates[:feature_subfeature_id] = action_definition[:feature_subfeature_id]
                end
                permission.assign_attributes(updates) if updates.size > 0
                permissions_to_keep.push(permission.id) if permission.id
              end
            end
          end
          component_permissions.where.not(id: permissions_to_keep).destroy_all
        end

        save if changed?

        RolePermission.where(component_permission: permissions_to_revoke.pluck(:id)).update_all(permitted: false) if permissions_to_revoke.any?
      end
    end
  end
end
